
from __future__ import absolute_import
#typing
from collections import defaultdict

#overrides

import torch
from torch.nn import Parameter
from torch.nn.modules.rnn import LSTMCell
from torch.nn.modules.linear import Linear

from allennlp.common import util as common_util
from allennlp.models.semantic_parsing.nlvr.nlvr_decoder_state import NlvrDecoderState
from allennlp.modules import Attention
from allennlp.nn.decoding import DecoderStep, RnnState, ChecklistState
from allennlp.nn import util as nn_util
try:
    from itertools import izip
except:
    izip = zip



class NlvrDecoderStep(DecoderStep[NlvrDecoderState]):
    u"""
    Parameters
    ----------
    encoder_output_dim : ``int``
    action_embedding_dim : ``int``
    input_attention : ``Attention``
    dropout : ``float``
        Dropout to use on decoder outputs and before action prediction.
    use_coverage : ``bool``, optional (default=False)
        Is this DecoderStep being used in a semantic parser trained using coverage? We need to know
        this to define a learned parameter for using checklist balances in action prediction.
    """
    def __init__(self,
                 encoder_output_dim     ,
                 action_embedding_dim     ,
                 input_attention           ,
                 dropout        = 0.0,
                 use_coverage       = False)        :
        super(NlvrDecoderStep, self).__init__()
        self._input_attention = input_attention

        # Decoder output dim needs to be the same as the encoder output dim since we initialize the
        # hidden state of the decoder with the final hidden state of the encoder.
        output_dim = encoder_output_dim
        input_dim = output_dim
        # Our decoder input will be the concatenation of the decoder hidden state and the previous
        # action embedding, and we'll project that down to the decoder's `input_dim`, which we
        # arbitrarily set to be the same as `output_dim`.
        self._input_projection_layer = Linear(output_dim + action_embedding_dim, input_dim)
        # Before making a prediction, we'll compute an attention over the input given our updated
        # hidden state. Then we concatenate those with the decoder state and project to
        # `action_embedding_dim` to make a prediction.
        self._output_projection_layer = Linear(output_dim + encoder_output_dim, action_embedding_dim)
        if use_coverage:
            # This is a multiplicative factor that is used to add the embeddings of yet to be
            # produced actions to the predicted embedding and bias it.
            self._checklist_embedding_multiplier = Parameter(torch.FloatTensor([1.0]))
        # TODO(pradeep): Do not hardcode decoder cell type.
        self._decoder_cell = LSTMCell(input_dim, output_dim)
        if dropout > 0:
            self._dropout = torch.nn.Dropout(p=dropout)
        else:
            self._dropout = lambda x: x

    #overrides
    def take_step(self,  # type: ignore
                  state                  ,
                  max_actions      = None,
                  allowed_actions                 = None)                          :
        u"""
        Given a ``NlvrDecoderState``, returns a list of next states that are sorted by their scores.
        This method is very similar to ``WikiTablesDecoderStep._take_step``. The differences are
        that depending on the type of supervision being used, we may not have a notion of
        "allowed actions" here, and we do not perform entity linking here.
        """
        # Outline here: first we'll construct the input to the decoder, which is a concatenation of
        # an embedding of the decoder input (the last action taken) and an attention over the
        # sentence.  Then we'll update our decoder's hidden state given this input, and recompute
        # an attention over the sentence given our new hidden state.  We'll use a concatenation of
        # the new hidden state, the new attention, and optionally the checklist balance to predict an
        # output, then yield new states. We will compute and use a checklist balance when
        # ``allowed_actions`` is None, with the assumption that the ``DecoderTrainer`` that is
        # calling this method is trying to train a parser without logical form supervision.
        # TODO (pradeep): Make the distinction between the two kinds of trainers in the way they
        # call this method more explicit.

        # Each new state corresponds to one valid action that can be taken from the current state,
        # and they are ordered by model scores.
        attended_sentence = torch.stack([rnn_state.attended_input for rnn_state in state.rnn_state])
        hidden_state = torch.stack([rnn_state.hidden_state for rnn_state in state.rnn_state])
        memory_cell = torch.stack([rnn_state.memory_cell for rnn_state in state.rnn_state])
        previous_action_embedding = torch.stack([rnn_state.previous_action_embedding
                                                 for rnn_state in state.rnn_state])

        # (group_size, decoder_input_dim)
        decoder_input = self._input_projection_layer(torch.cat([attended_sentence,
                                                                previous_action_embedding], -1))
        decoder_input = torch.tanh(decoder_input)
        hidden_state, memory_cell = self._decoder_cell(decoder_input, (hidden_state, memory_cell))

        hidden_state = self._dropout(hidden_state)
        # (group_size, encoder_output_dim)
        encoder_outputs = torch.stack([state.rnn_state[0].encoder_outputs[i] for i in state.batch_indices])
        encoder_output_mask = torch.stack([state.rnn_state[0].encoder_output_mask[i] for i in state.batch_indices])
        attended_sentence = self.attend_on_sentence(hidden_state, encoder_outputs, encoder_output_mask)

        # We get global indices of actions to embed here. The following logic is similar to
        # ``WikiTablesDecoderStep._get_actions_to_consider``, except that we do not have any actions
        # to link.
        valid_actions = state.get_valid_actions()
        global_valid_actions                              = []
        for batch_index, valid_action_list in izip(state.batch_indices, valid_actions):
            global_valid_actions.append([])
            for action_index in valid_action_list:
                # state.action_indices is a dictionary that maps (batch_index, batch_action_index)
                # to global_action_index
                global_action_index = state.action_indices[(batch_index, action_index)]
                global_valid_actions[-1].append((global_action_index, action_index))
        global_actions_to_embed                  = []
        local_actions                  = []
        for global_action_list in global_valid_actions:
            global_action_list.sort()
            global_actions_to_embed.append([])
            local_actions.append([])
            for global_action_index, action_index in global_action_list:
                global_actions_to_embed[-1].append(global_action_index)
                local_actions[-1].append(action_index)
        max_num_actions = max([len(action_list) for action_list in global_actions_to_embed])
        # We pad local actions with -1 as padding to get considered actions.
        considered_actions = [common_util.pad_sequence_to_length(action_list, max_num_actions,
                                                                 default_value=lambda: -1)
                              for action_list in local_actions]

        # action_embeddings: (group_size, num_embedded_actions, action_embedding_dim)
        # action_mask: (group_size, num_embedded_actions)
        action_embeddings, embedded_action_mask = self._get_action_embeddings(state,
                                                                              global_actions_to_embed)
        action_query = torch.cat([hidden_state, attended_sentence], dim=-1)
        # (group_size, action_embedding_dim)
        predicted_action_embedding = self._output_projection_layer(action_query)
        predicted_action_embedding = self._dropout(torch.tanh(predicted_action_embedding))
        if state.checklist_state[0] is not None:
            embedding_addition = self._get_predicted_embedding_addition(state)
            addition = embedding_addition * self._checklist_embedding_multiplier
            predicted_action_embedding = predicted_action_embedding + addition
        # We'll do a batch dot product here with `bmm`.  We want `dot(predicted_action_embedding,
        # action_embedding)` for each `action_embedding`, and we can get that efficiently with
        # `bmm` and some squeezing.
        # Shape: (group_size, num_embedded_actions)
        action_logits = action_embeddings.bmm(predicted_action_embedding.unsqueeze(-1)).squeeze(-1)

        action_mask = embedded_action_mask.float()
        if state.checklist_state[0] is not None:
            # We will compute the logprobs and the checklists of potential next states together for
            # efficiency.
            logprobs, new_checklist_states = self._get_next_state_info_with_agenda(state,
                                                                                   considered_actions,
                                                                                   action_logits,
                                                                                   action_mask)
        else:
            logprobs = self._get_next_state_info_without_agenda(state,
                                                                considered_actions,
                                                                action_logits,
                                                                action_mask)
            new_checklist_states = None
        return self._compute_new_states(state,
                                        logprobs,
                                        hidden_state,
                                        memory_cell,
                                        action_embeddings,
                                        attended_sentence,
                                        considered_actions,
                                        allowed_actions,
                                        new_checklist_states,
                                        max_actions)

    @staticmethod
    def _get_predicted_embedding_addition(state                  )                :
        u"""
        Computes checklist balance, uses it to get the embeddings of desired terminal actions yet to
        be produced by the decoder, and returns their sum for the decoder to add it to the predicted
        embedding to bias the prediction towards missing actions.
        """
        # This holds a list of checklist balances for this state. Each balance is a float vector
        # containing just 1s and 0s showing which of the items are filled. We clamp the min at 0
        # to ignore the number of times an action is taken. The value at an index will be 1 iff
        # the target wants an unmasked action to be taken, and it is not yet taken. All elements
        # in each balance corresponding to masked actions will be 0.
        checklist_balances = []
        for instance_checklist_state in state.checklist_state:
            checklist_balances.append(instance_checklist_state.get_balance())

        # Note that the checklist balance has 0s for actions that are masked (see comment above).
        # So, masked actions do not contribute to the projection of ``predicted_action_emebedding``
        # below.
        # (group_size, num_terminals, 1)
        checklist_balance = torch.stack([x for x in  checklist_balances])
        global_terminal_indices                  = []
        # We map the group indices of all terminal actions that are relevant to checklist
        # computation, to global indices here.
        for batch_index, checklist_state in izip(state.batch_indices, state.checklist_state):
            global_terminal_indices.append([])
            for terminal_index in checklist_state.terminal_actions.detach().cpu():
                global_terminal_index = state.action_indices[(batch_index, int(terminal_index[0]))]
                global_terminal_indices[-1].append(global_terminal_index)
        # We don't need to pad this tensor because the terminal indices from all groups will be the
        # same size.
        terminal_indices_tensor = state.score[0].new_tensor(global_terminal_indices, dtype=torch.long)
        group_size = len(state.batch_indices)
        action_embedding_dim = state.action_embeddings.size(-1)
        num_terminals = len(global_terminal_indices[0])
        flattened_terminal_indices = terminal_indices_tensor.view(-1)
        flattened_action_embeddings = state.action_embeddings.index_select(0,
                                                                           flattened_terminal_indices)
        terminal_embeddings = flattened_action_embeddings.view(group_size, num_terminals, action_embedding_dim)
        checklist_balance_embeddings = terminal_embeddings * checklist_balance
        # (group_size, action_embedding_dim)
        return checklist_balance_embeddings.sum(1)

    @staticmethod
    def _get_next_state_info_with_agenda(
            state                  ,
            considered_actions                 ,
            action_logits              ,
            action_mask              ):                                                   
                                                                           
        u"""
        We return a list of log probabilities and checklist states corresponding to next actions that are
        not padding. This method is applicable to the case where we do not have target action
        sequences and are relying on agendas for training.
        """
        considered_action_probs = nn_util.masked_softmax(action_logits, action_mask)
        # Mixing model scores and agenda selection probabilities to compute the probabilities of all
        # actions for the next step and the corresponding new checklists.
        # All action logprobs will keep track of logprob corresponding to each local action index
        # for each instance.
        all_action_logprobs                                           = []
        all_new_checklist_states                             = []
        for group_index, instance_info in enumerate(izip(state.score,
                                                        considered_action_probs,
                                                        state.checklist_state)):
            (instance_score, instance_probs, instance_checklist_state) = instance_info
            # We will mix the model scores with agenda selection probabilities and compute their
            # logs to fill the following list with action indices and corresponding logprobs.
            instance_action_logprobs                                 = []
            instance_new_checklist_states                       = []
            for action_index, action_prob in enumerate(instance_probs):
                # This is the actual index of the action from the original list of actions.
                action = considered_actions[group_index][action_index]
                if action == -1:
                    # Ignoring padding.
                    continue
                new_checklist_state = instance_checklist_state.update(action)  # (terminal_actions, 1)
                instance_new_checklist_states.append(new_checklist_state)
                logprob = instance_score + torch.log(action_prob + 1e-13)
                instance_action_logprobs.append((action_index, logprob))
            all_action_logprobs.append(instance_action_logprobs)
            all_new_checklist_states.append(instance_new_checklist_states)
        return all_action_logprobs, all_new_checklist_states

    @staticmethod
    def _get_next_state_info_without_agenda(state                  ,
                                            considered_actions                 ,
                                            action_logits              ,
                                            action_mask              ):                        
                                                                                                             
        u"""
        We return a list of log probabilities corresponding to actions that are not padding. This
        method is related to the training scenario where we have target action sequences for
        training.
        """
        considered_action_logprobs = nn_util.masked_log_softmax(action_logits, action_mask)
        all_action_logprobs                                           = []
        for group_index, (score, considered_logprobs) in enumerate(izip(state.score,
                                                                       considered_action_logprobs)):
            instance_action_logprobs                                 = []
            for action_index, logprob in enumerate(considered_logprobs):
                # This is the actual index of the action from the original list of actions.
                action = considered_actions[group_index][action_index]
                if action == -1:
                    # Ignoring padding.
                    continue
                instance_action_logprobs.append((action_index, score + logprob))
            all_action_logprobs.append(instance_action_logprobs)
        return all_action_logprobs

    def attend_on_sentence(self,
                           query              ,
                           encoder_outputs              ,
                           encoder_output_mask              )                                     :
        u"""
        This method is almost identical to ``WikiTablesDecoderStep.attend_on_question``. We just
        don't return the attention weights.
        Given a query (which is typically the decoder hidden state), compute an attention over the
        output of the sentence encoder, and return a weighted sum of the sentence representations
        given this attention.  We also return the attention weights themselves.

        This is a simple computation, but we have it as a separate method so that the ``forward``
        method on the main parser module can call it on the initial hidden state, to simplify the
        logic in ``take_step``.
        """
        # (group_size, sentence_length)
        sentence_attention_weights = self._input_attention(query,
                                                           encoder_outputs,
                                                           encoder_output_mask)
        # (group_size, encoder_output_dim)
        attended_sentence = nn_util.weighted_sum(encoder_outputs, sentence_attention_weights)
        return attended_sentence

    @staticmethod
    def _get_action_embeddings(state                  ,
                               actions_to_embed                 )                                     :
        u"""
        This method is identical to ``WikiTablesDecoderStep._get_action_embeddings``
        Returns an embedded representation for all actions in ``actions_to_embed``, using the state
        in ``NlvrDecoderState``.

        Parameters
        ----------
        state : ``NlvrDecoderState``
            The current state.  We'll use this to get the global action embeddings.
        actions_to_embed : ``List[List[int]]``
            A list of _global_ action indices for each group element.  Should have shape
            (group_size, num_actions), unpadded.

        Returns
        -------
        action_embeddings : ``torch.FloatTensor``
            An embedded representation of all of the given actions.  Shape is ``(group_size,
            num_actions, action_embedding_dim)``, where ``num_actions`` is the maximum number of
            considered actions for any group element.
        action_mask : ``torch.LongTensor``
            A mask of shape ``(group_size, num_actions)`` indicating which ``(group_index,
            action_index)`` pairs were merely added as padding.
        """
        num_actions = [len(action_list) for action_list in actions_to_embed]
        max_num_actions = max(num_actions)
        padded_actions = [common_util.pad_sequence_to_length(action_list, max_num_actions)
                          for action_list in actions_to_embed]
        # Shape: (group_size, num_actions)
        action_tensor = state.score[0].new_tensor(padded_actions, dtype=torch.long)
        # `state.action_embeddings` is shape (total_num_actions, action_embedding_dim).
        # We want to select from state.action_embeddings using `action_tensor` to get a tensor of
        # shape (group_size, num_actions, action_embedding_dim).  Unfortunately, the index_select
        # functions in nn.util don't do this operation.  So we'll do some reshapes and do the
        # index_select ourselves.
        group_size = len(state.batch_indices)
        action_embedding_dim = state.action_embeddings.size(-1)
        flattened_actions = action_tensor.view(-1)
        flattened_action_embeddings = state.action_embeddings.index_select(0, flattened_actions)
        action_embeddings = flattened_action_embeddings.view(group_size, max_num_actions, action_embedding_dim)
        sequence_lengths = action_embeddings.new_tensor(num_actions)
        action_mask = nn_util.get_mask_from_sequence_lengths(sequence_lengths, max_num_actions)
        return action_embeddings, action_mask

    @classmethod
    def _compute_new_states(cls,
                            state                  ,
                            action_logprobs                                      ,
                            hidden_state              ,
                            memory_cell              ,
                            action_embeddings              ,
                            attended_sentence              ,
                            considered_actions                 ,
                            allowed_actions                 = None,
                            new_checklist_states                             = None,
                            max_actions      = None)                          :
        u"""
        This method is very similar to ``WikiTabledDecoderStep._compute_new_states``.
        The difference here is that we also keep track of checklists if they are passed to this
        method.
        """
        # batch_index -> group_index, action_index, checklist, score
        states_info                                                               = defaultdict(list)
        if new_checklist_states is None:
            # We do not have checklist states. Making a list of lists of Nones of the appropriate size for
            # the zips below.
            new_checklist_states = [[None for logprob in instance_logprobs]
                                    for instance_logprobs in action_logprobs]
        for group_index, instance_info in enumerate(izip(state.batch_indices,
                                                        action_logprobs,
                                                        new_checklist_states)):
            batch_index, instance_logprobs, instance_new_checklist_states = instance_info
            for (action_index, score), checklist_state in izip(instance_logprobs,
                                                              instance_new_checklist_states):
                states_info[batch_index].append((group_index, action_index, checklist_state, score))

        new_states = []
        for batch_index, instance_states_info in list(states_info.items()):
            batch_scores = torch.cat([info[-1] for info in instance_states_info])
            _, sorted_indices = batch_scores.sort(-1, descending=True)
            sorted_states_info = [instance_states_info[i] for i in sorted_indices.detach().cpu().numpy()]
            allowed_states_info = []
            for i, (group_index, action_index, _, _) in enumerate(sorted_states_info):
                action = considered_actions[group_index][action_index]
                if allowed_actions is not None and action not in allowed_actions[group_index]:
                    continue
                allowed_states_info.append(sorted_states_info[i])
            sorted_states_info = allowed_states_info
            if max_actions is not None:
                sorted_states_info = sorted_states_info[:max_actions]
            for group_index, action_index, new_checklist_state, new_score in sorted_states_info:
                # This is the actual index of the action from the original list of actions.
                # We do not have to check whether it is the padding index because ``take_step``
                # already took care of that.
                action = considered_actions[group_index][action_index]
                action_embedding = action_embeddings[group_index, action_index, :]
                new_action_history = state.action_history[group_index] + [action]
                production_rule = state.possible_actions[batch_index][action][0]
                new_grammar_state = state.grammar_state[group_index].take_action(production_rule)
                new_rnn_state = RnnState(hidden_state[group_index],
                                         memory_cell[group_index],
                                         action_embedding,
                                         attended_sentence[group_index],
                                         state.rnn_state[group_index].encoder_outputs,
                                         state.rnn_state[group_index].encoder_output_mask)
                new_state = NlvrDecoderState(batch_indices=[batch_index],
                                             action_history=[new_action_history],
                                             score=[new_score],
                                             rnn_state=[new_rnn_state],
                                             grammar_state=[new_grammar_state],
                                             action_embeddings=state.action_embeddings,
                                             action_indices=state.action_indices,
                                             possible_actions=state.possible_actions,
                                             worlds=state.worlds,
                                             label_strings=state.label_strings,
                                             checklist_state=[new_checklist_state])
                new_states.append(new_state)
        return new_states




from __future__ import absolute_import
#typing

from allennlp.common import Registrable
from allennlp.data.tokenizers.token import Token


class Tokenizer(Registrable):
    u"""
    A ``Tokenizer`` splits strings of text into tokens.  Typically, this either splits text into
    word tokens or character tokens, and those are the two tokenizer subclasses we have implemented
    here, though you could imagine wanting to do other kinds of tokenization for structured or
    other inputs.

    As part of tokenization, concrete implementations of this API will also handle stemming,
    stopword filtering, adding start and end tokens, or other kinds of things you might want to do
    to your tokens.  See the parameters to, e.g., :class:`~.WordTokenizer`, or whichever tokenizer
    you want to use.

    If the base input to your model is words, you should use a :class:`~.WordTokenizer`, even if
    you also want to have a character-level encoder to get an additional vector for each word
    token.  Splitting word tokens into character arrays is handled separately, in the
    :class:`..token_representations.TokenRepresentation` class.
    """
    default_implementation = u'word'

    def batch_tokenize(self, texts           )                     :
        u"""
        Batches together tokenization of several texts, in case that is faster for particular
        tokenizers.
        """
        raise NotImplementedError

    def tokenize(self, text     )               :
        u"""
        Actually implements splitting words into tokens.

        Returns
        -------
        tokens : ``List[Token]``
        """
        raise NotImplementedError



# pylint: disable=no-self-use,invalid-name,line-too-long



from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import json
import os

import flask
import flask.testing

from allennlp.common.util import JsonDict
from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor
from allennlp.service.server_simple import make_app
from io import open


def post_json(client                           , endpoint     , data          )                  :
    return client.post(endpoint,
                       content_type=u"application/json",
                       data=json.dumps(data))

PAYLOAD = {
        u'passage': u"""The Matrix is a 1999 science fiction action film written and directed by The Wachowskis, starring Keanu Reeves, Laurence Fishburne, Carrie-Anne Moss, Hugo Weaving, and Joe Pantoliano.""",
        u'question': u"""Who stars in the matrix?"""
}


class TestSimpleServer(AllenNlpTestCase):

    def setUp(self):
        super(TestSimpleServer, self).setUp()

        archive = load_archive(self.FIXTURES_ROOT / u'bidaf' / u'serialization' / u'model.tar.gz')
        self.bidaf_predictor = Predictor.from_archive(archive, u'machine-comprehension')


    def tearDown(self):
        super(TestSimpleServer, self).tearDown()
        try:
            os.remove(u'access.log')
            os.remove(u'error.log')
        except FileNotFoundError:
            pass

    def test_standard_model(self):
        app = make_app(predictor=self.bidaf_predictor, field_names=[u'passage', u'question'])
        app.testing = True
        client = app.test_client()

        # First test the HTML
        response = client.get(u'/')
        data = response.get_data()

        assert u"passage" in data
        assert u"question" in data

        # Now test the backend
        response = post_json(client, u'/predict', PAYLOAD)
        data = json.loads(response.get_data())
        assert u'best_span_str' in data
        assert u'span_start_logits' in data

    def test_sanitizer(self):
        def sanitize(result          )            :
            return dict((key, value) for key, value in list(result.items())
                    if key.startswith(u"best_span"))

        app = make_app(predictor=self.bidaf_predictor, field_names=[u'passage', u'question'], sanitizer=sanitize)
        app.testing = True
        client = app.test_client()

        response = post_json(client, u'/predict', PAYLOAD)
        data = json.loads(response.get_data())
        assert u'best_span_str' in data
        assert u'span_start_logits' not in data

    def test_static_dir(self):
        html = u"""<html><body>THIS IS A STATIC SITE</body></html>"""
        jpg = u"""something about a jpg"""

        with open(os.path.join(self.TEST_DIR, u'index.html'), u'w') as f:
            f.write(html)

        with open(os.path.join(self.TEST_DIR, u'jpg.txt'), u'w') as f:
            f.write(jpg)

        app = make_app(predictor=self.bidaf_predictor, static_dir=self.TEST_DIR)
        app.testing = True
        client = app.test_client()

        response = client.get(u'/')
        data = response.get_data().decode(u'utf-8')
        assert data == html

        response = client.get(u'jpg.txt')
        data = response.get_data().decode(u'utf-8')
        assert data == jpg



from __future__ import absolute_import
#typing

import torch
from allennlp.data.fields.production_rule_field import ProductionRuleArray
from allennlp.nn.decoding import DecoderState, GrammarState, RnnState, ChecklistState
from allennlp.semparse.worlds import NlvrWorld
try:
    from itertools import izip
except:
    izip = zip


class NlvrDecoderState(DecoderState[u'NlvrDecoderState']):
    u"""
    This class is very similar to ``WikiTablesDecoderState``, except that we keep track of a
    checklist score, and other variables related to it.

    Parameters
    ----------
    batch_indices : ``List[int]``
        Passed to super class; see docs there.
    action_history : ``List[List[int]]``
        Passed to super class; see docs there.
    score : ``List[torch.Tensor]``
        Passed to super class; see docs there.
    rnn_state : ``List[RnnState]``
        An ``RnnState`` for every group element.  This keeps track of the current decoder hidden
        state, the previous decoder output, the output from the encoder (for computing attentions),
        and other things that are typical seq2seq decoder state things.
    grammar_state : ``List[GrammarState]``
        This hold the current grammar state for each element of the group.  The ``GrammarState``
        keeps track of which actions are currently valid.
    action_embeddings : ``torch.Tensor``
        The global action embeddings tensor.  Has shape ``(num_global_embeddable_actions,
        action_embedding_dim)``.
    action_indices : ``Dict[Tuple[int, int], int]``
        A mapping from ``(batch_index, action_index)`` to ``global_action_index``.
    possible_actions : ``List[List[ProductionRuleArray]]``
        The list of all possible actions that was passed to ``model.forward()``.  We need this so
        we can recover production strings, which we need to update grammar states.
    worlds : ``List[List[NlvrWorld]]``
        The worlds associated with each element. These are needed to compute the denotations. The
        outer list corresponds to elements, and the inner corresponds to worlds related to each
        element.
    label_strings : ``List[List[str]]``
        String representations of labels for the elements provided. When scoring finished states, we
        will compare the denotations of their action sequences against these labels. For each
        element, there will be as many labels as there are worlds.
    checklist_state : ``List[ChecklistState]``, optional (default=None)
        If you are using this state within a parser being trained for coverage, we need to store a
        ``ChecklistState`` which keeps track of the coverage information. Not needed if you are
        using a non-coverage based training algorithm.
    """
    def __init__(self,
                 batch_indices           ,
                 action_history                 ,
                 score                    ,
                 rnn_state                ,
                 grammar_state                    ,
                 action_embeddings              ,
                 action_indices                            ,
                 possible_actions                                 ,
                 worlds                       ,
                 label_strings                 ,
                 checklist_state                       = None)        :
        super(NlvrDecoderState, self).__init__(batch_indices, action_history, score)
        self.rnn_state = rnn_state
        self.grammar_state = grammar_state
        # Converting None to list of Nones if needed, to simplify state operations.
        self.checklist_state = checklist_state if checklist_state is not None else [None for _ in
                                                                                    batch_indices]
        self.action_embeddings = action_embeddings
        self.action_indices = action_indices
        self.possible_actions = possible_actions
        self.worlds = worlds
        self.label_strings = label_strings

    def print_action_history(self, group_index      = None)        :
        scores = self.score if group_index is None else [self.score[group_index]]
        batch_indices = self.batch_indices if group_index is None else [self.batch_indices[group_index]]
        histories = self.action_history if group_index is None else [self.action_history[group_index]]
        for score, batch_index, action_history in izip(scores, batch_indices, histories):
            print u'  ', score.detach().cpu().numpy()[0],
                  [self.possible_actions[batch_index][action][0] for action in action_history]

    def get_valid_actions(self)                   :
        u"""
        Returns a list of valid actions for each element of the group.
        """
        valid_actions = [state.get_valid_actions() for state in self.grammar_state]
        return valid_actions

    def is_finished(self)        :
        u"""This method is identical to ``WikiTablesDecoderState.is_finished``."""
        if len(self.batch_indices) != 1:
            raise RuntimeError(u"is_finished() is only defined with a group_size of 1")
        return self.grammar_state[0].is_finished()

    @classmethod
    def combine_states(cls, states)                      :
        batch_indices = [batch_index for state in states for batch_index in state.batch_indices]
        action_histories = [action_history for state in states for action_history in state.action_history]
        scores = [score for state in states for score in state.score]
        rnn_states = [rnn_state for state in states for rnn_state in state.rnn_state]
        grammar_states = [grammar_state for state in states for grammar_state in state.grammar_state]
        checklist_states = [checklist_state for state in states for checklist_state in state.checklist_state]
        return NlvrDecoderState(batch_indices=batch_indices,
                                action_history=action_histories,
                                score=scores,
                                rnn_state=rnn_states,
                                grammar_state=grammar_states,
                                action_embeddings=states[0].action_embeddings,
                                action_indices=states[0].action_indices,
                                possible_actions=states[0].possible_actions,
                                worlds=states[0].worlds,
                                label_strings=states[0].label_strings,
                                checklist_state=checklist_states)



# pylint: disable=invalid-name

from __future__ import absolute_import
import os
import torch
from torch.utils.ffi import create_extension

if not torch.cuda.is_available():
    raise Exception(u'HighwayLSTM can only be compiled with CUDA')

sources = [u'src/highway_lstm_cuda.c']
headers = [u'src/highway_lstm_cuda.h']
defines = [(u'WITH_CUDA', None)]
with_cuda = True

this_file = os.path.dirname(os.path.realpath(__file__))
extra_objects = [u'src/highway_lstm_kernel.cu.o']
extra_objects = [os.path.join(this_file, fname) for fname in extra_objects]

ffi = create_extension(
        u'_ext.highway_lstm_layer',
        headers=headers,
        sources=sources,
        define_macros=defines,
        relative_to=__file__,
        with_cuda=with_cuda,
        extra_objects=extra_objects
        )

if __name__ == u'__main__':
    ffi.build()



u"""
A :class:`~allennlp.training.trainer.Trainer` is responsible for training a
:class:`~allennlp.models.model.Model`.

Typically you might create a configuration file specifying the model and
training parameters and then use :mod:`~allennlp.commands.train`
rather than instantiating a ``Trainer`` yourself.
"""
# pylint: disable=too-many-lines




from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
import logging
import os
import shutil
import time
import re
import datetime
import traceback
#typing

import torch
import torch.optim.lr_scheduler
from torch.nn.parallel import replicate, parallel_apply
from torch.nn.parallel.scatter_gather import scatter_kwargs, gather
SummaryWriter = None


from allennlp.common import Params
from allennlp.common.checks import ConfigurationError
from allennlp.common.util import peak_memory_mb, gpu_memory_mb
from allennlp.common.tqdm import Tqdm
from allennlp.data.instance import Instance
from allennlp.data.iterators.data_iterator import DataIterator
from allennlp.models.model import Model
from allennlp.nn import util
from allennlp.training.learning_rate_schedulers import LearningRateScheduler
from allennlp.training.optimizers import Optimizer
try:
    from itertools import ifilter
except:
    ifilter = filter

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def is_sparse(tensor):
    return tensor.is_sparse


def sparse_clip_norm(parameters, max_norm, norm_type=2)         :
    u"""Clips gradient norm of an iterable of parameters.

    The norm is computed over all gradients together, as if they were
    concatenated into a single vector. Gradients are modified in-place.
    Supports sparse gradients.

    Parameters
    ----------
    parameters : ``(Iterable[torch.Tensor])``
        An iterable of Tensors that will have gradients normalized.
    max_norm : ``float``
        The max norm of the gradients.
    norm_type : ``float``
        The type of the used p-norm. Can be ``'inf'`` for infinity norm.

    Returns
    -------
    Total norm of the parameters (viewed as a single vector).
    """
    # pylint: disable=invalid-name,protected-access
    parameters = list(ifilter(lambda p: p.grad is not None, parameters))
    max_norm = float(max_norm)
    norm_type = float(norm_type)
    if norm_type == float(u'inf'):
        total_norm = max(p.grad.data.abs().max() for p in parameters)
    else:
        total_norm = 0
        for p in parameters:
            if is_sparse(p.grad):
                # need to coalesce the repeated indices before finding norm
                grad = p.grad.data.coalesce()
                param_norm = grad._values().norm(norm_type)
            else:
                param_norm = p.grad.data.norm(norm_type)
            total_norm += param_norm ** norm_type
        total_norm = total_norm ** (1. / norm_type)
    clip_coef = max_norm / (total_norm + 1e-6)
    if clip_coef < 1:
        for p in parameters:
            if is_sparse(p.grad):
                p.grad.data._values().mul_(clip_coef)
            else:
                p.grad.data.mul_(clip_coef)
    return total_norm


def move_optimizer_to_cuda(optimizer):
    u"""
    Move the optimizer state to GPU, if necessary.
    After calling, any parameter specific state in the optimizer
    will be located on the same device as the parameter.
    """
    for param_group in optimizer.param_groups:
        for param in param_group[u'params']:
            if param.is_cuda:
                param_state = optimizer.state[param]
                for k in list(param_state.keys()):
                    if isinstance(param_state[k], torch.Tensor):
                        param_state[k] = param_state[k].cuda(device=param.get_device())


class TensorboardWriter(object):
    u"""
    Wraps a pair of ``SummaryWriter`` instances but is a no-op if they're ``None``.
    Allows Tensorboard logging without always checking for Nones first.
    """
    def __init__(self, train_log                = None, validation_log                = None)        :
        self._train_log = train_log
        self._validation_log = validation_log

    @staticmethod
    def _item(value     ):
        if hasattr(value, u'item'):
            val = value.item()
        else:
            val = value
        return val

    def add_train_scalar(self, name     , value       , global_step     )        :
        # get the scalar
        if self._train_log is not None:
            self._train_log.add_scalar(name, self._item(value), global_step)

    def add_train_histogram(self, name     , values              , global_step     )        :
        if self._train_log is not None:
            if isinstance(values, torch.Tensor):
                values_to_write = values.cpu().data.numpy().flatten()
                self._train_log.add_histogram(name, values_to_write, global_step)

    def add_validation_scalar(self, name     , value       , global_step     )        :

        if self._validation_log is not None:
            self._validation_log.add_scalar(name, self._item(value), global_step)


def time_to_str(timestamp     )       :
    u"""
    Convert seconds past Epoch to human readable string.
    """
    datetimestamp = datetime.datetime.fromtimestamp(timestamp)
    return u'{:04d}-{:02d}-{:02d}-{:02d}-{:02d}-{:02d}'.format(
            datetimestamp.year, datetimestamp.month, datetimestamp.day,
            datetimestamp.hour, datetimestamp.minute, datetimestamp.second
    )


def str_to_time(time_str     )                     :
    u"""
    Convert human readable string to datetime.datetime.
    """
    pieces      = [int(piece) for piece in time_str.split(u'-')]
    return datetime.datetime(*pieces)


class Trainer(object):
    def __init__(self,
                 model       ,
                 optimizer                       ,
                 iterator              ,
                 train_dataset                    ,
                 validation_dataset                               = None,
                 patience                = None,
                 validation_metric      = u"-loss",
                 validation_iterator               = None,
                 shuffle       = True,
                 num_epochs      = 20,
                 serialization_dir                = None,
                 num_serialized_models_to_keep      = 20,
                 keep_serialized_model_every_num_seconds      = None,
                 model_save_interval        = None,
                 cuda_device                   = -1,
                 grad_norm                  = None,
                 grad_clipping                  = None,
                 learning_rate_scheduler                                  = None,
                 summary_interval      = 100,
                 histogram_interval      = None)        :
        u"""
        Parameters
        ----------
        model : ``Model``, required.
            An AllenNLP model to be optimized. Pytorch Modules can also be optimized if
            their ``forward`` method returns a dictionary with a "loss" key, containing a
            scalar tensor representing the loss function to be optimized.
        optimizer : ``torch.nn.Optimizer``, required.
            An instance of a Pytorch Optimizer, instantiated with the parameters of the
            model to be optimized.
        iterator : ``DataIterator``, required.
            A method for iterating over a ``Dataset``, yielding padded indexed batches.
        train_dataset : ``Dataset``, required.
            A ``Dataset`` to train on. The dataset should have already been indexed.
        validation_dataset : ``Dataset``, optional, (default = None).
            A ``Dataset`` to evaluate on. The dataset should have already been indexed.
        patience : Optional[int] > 0, optional (default=None)
            Number of epochs to be patient before early stopping: the training is stopped
            after ``patience`` epochs with no improvement. If given, it must be ``> 0``.
            If None, early stopping is disabled.
        validation_metric : str, optional (default="loss")
            Validation metric to measure for whether to stop training using patience
            and whether to serialize an ``is_best`` model each epoch. The metric name
            must be prepended with either "+" or "-", which specifies whether the metric
            is an increasing or decreasing function.
        validation_iterator : ``DataIterator``, optional (default=None)
            An iterator to use for the validation set.  If ``None``, then
            use the training `iterator`.
        shuffle: ``bool``, optional (default=True)
            Whether to shuffle the instances in the iterator or not.
        num_epochs : int, optional (default = 20)
            Number of training epochs.
        serialization_dir : str, optional (default=None)
            Path to directory for saving and loading model files. Models will not be saved if
            this parameter is not passed.
        num_serialized_models_to_keep : ``int``, optional (default=20)
            Number of previous model checkpoints to retain.  Default is to keep 20 checkpoints.
            A value of None or -1 means all checkpoints will be kept.
        keep_serialized_model_every_num_seconds : ``int``, optional (default=None)
            If num_serialized_models_to_keep is not None, then occasionally it's useful to
            save models at a given interval in addition to the last num_serialized_models_to_keep.
            To do so, specify keep_serialized_model_every_num_seconds as the number of seconds
            between permanently saved checkpoints.  Note that this option is only used if
            num_serialized_models_to_keep is not None, otherwise all checkpoints are kept.
        model_save_interval : ``float``, optional (default=None)
            If provided, then serialize models every ``model_save_interval``
            seconds within single epochs.  In all cases, models are also saved
            at the end of every epoch if ``serialization_dir`` is provided.
        cuda_device : ``int``, optional (default = -1)
            An integer specifying the CUDA device to use. If -1, the CPU is used.
        grad_norm : ``float``, optional, (default = None).
            If provided, gradient norms will be rescaled to have a maximum of this value.
        grad_clipping : ``float``, optional (default = ``None``).
            If provided, gradients will be clipped `during the backward pass` to have an (absolute)
            maximum of this value.  If you are getting ``NaNs`` in your gradients during training
            that are not solved by using ``grad_norm``, you may need this.
        learning_rate_scheduler : ``PytorchLRScheduler``, optional, (default = None)
            A Pytorch learning rate scheduler. The learning rate will be decayed with respect to
            this schedule at the end of each epoch. If you use
            :class:`torch.optim.lr_scheduler.ReduceLROnPlateau`, this will use the ``validation_metric``
            provided to determine if learning has plateaued.  To support updating the learning
            rate on every batch, this can optionally implement ``step_batch(batch_num_total)`` which
            updates the learning rate given the batch number.
        summary_interval: ``int``, optional, (default = 100)
            Number of batches between logging scalars to tensorboard
        histogram_interval : ``int``, optional, (default = ``None``)
            If not None, then log histograms to tensorboard every ``histogram_interval`` batches.
            When this parameter is specified, the following additional logging is enabled:
                * Histograms of model parameters
                * The ratio of parameter update norm to parameter norm
                * Histogram of layer activations
            We log histograms of the parameters returned by
            ``model.get_parameters_for_histogram_tensorboard_logging``.
            The layer activations are logged for any modules in the ``Model`` that have
            the attribute ``should_log_activations`` set to ``True``.  Logging
            histograms requires a number of GPU-CPU copies during training and is typically
            slow, so we recommend logging histograms relatively infrequently.
            Note: only Modules that return tensors, tuples of tensors or dicts
            with tensors as values currently support activation logging.
        """
        self._model = model
        self._iterator = iterator
        self._validation_iterator = validation_iterator
        self._shuffle = shuffle
        self._optimizer = optimizer
        self._train_data = train_dataset
        self._validation_data = validation_dataset

        if patience is None:  # no early stopping
            if validation_dataset:
                logger.warning(u'You provided a validation dataset but patience was set to None, '
                               u'meaning that early stopping is disabled')
        elif (not isinstance(patience, int)) or patience <= 0:
            raise ConfigurationError(u'{} is an invalid value for "patience": it must be a positive integer '
                                     u'or None (if you want to disable early stopping)'.format(patience))
        self._patience = patience
        self._num_epochs = num_epochs

        self._serialization_dir = serialization_dir
        self._num_serialized_models_to_keep = num_serialized_models_to_keep
        self._keep_serialized_model_every_num_seconds = keep_serialized_model_every_num_seconds
        self._serialized_paths = []
        self._last_permanent_saved_checkpoint_time = time.time()
        self._model_save_interval = model_save_interval

        self._grad_norm = grad_norm
        self._grad_clipping = grad_clipping
        self._learning_rate_scheduler = learning_rate_scheduler

        increase_or_decrease = validation_metric[0]
        if increase_or_decrease not in [u"+", u"-"]:
            raise ConfigurationError(u"Validation metrics must specify whether they should increase "
                                     u"or decrease by pre-pending the metric name with a +/-.")
        self._validation_metric = validation_metric[1:]
        self._validation_metric_decreases = increase_or_decrease == u"-"

        if not isinstance(cuda_device, int) and not isinstance(cuda_device, list):
            raise ConfigurationError(u"Expected an int or list for cuda_device, got {}".format(cuda_device))

        if isinstance(cuda_device, list):
            logger.info("WARNING: Multiple GPU support is experimental not recommended for use. "
                        u"In some cases it may lead to incorrect results or undefined behavior.")
            self._multiple_gpu = True
            self._cuda_devices = cuda_device
            # data_parallel will take care of transfering to cuda devices,
            # so the iterator keeps data on CPU.
            self._iterator_device = -1
        else:
            self._multiple_gpu = False
            self._cuda_devices = [cuda_device]
            self._iterator_device = cuda_device

        if self._cuda_devices[0] != -1:
            self._model = self._model.cuda(self._cuda_devices[0])

        self._log_interval = 10  # seconds
        self._summary_interval = summary_interval
        self._histogram_interval = histogram_interval
        self._log_histograms_this_batch = False
        # We keep the total batch number as a class variable because it
        # is used inside a closure for the hook which logs activations in
        # ``_enable_activation_logging``.
        self._batch_num_total = 0

        self._last_log = 0.0  # time of last logging

        if serialization_dir is not None:
            train_log = SummaryWriter(os.path.join(serialization_dir, u"log", u"train"))
            validation_log = SummaryWriter(os.path.join(serialization_dir, u"log", u"validation"))
            self._tensorboard = TensorboardWriter(train_log, validation_log)
        else:
            self._tensorboard = TensorboardWriter()
        self._warned_tqdm_ignores_underscores = False

    def _enable_gradient_clipping(self)        :
        if self._grad_clipping is not None:
            # Pylint is unable to tell that we're in the case that _grad_clipping is not None...
            # pylint: disable=invalid-unary-operand-type
            clip_function = lambda grad: grad.clamp(-self._grad_clipping, self._grad_clipping)
            for parameter in self._model.parameters():
                if parameter.requires_grad:
                    parameter.register_hook(clip_function)

    def _enable_activation_logging(self)        :
        u"""
        Log activations to tensorboard
        """
        if self._histogram_interval is not None:
            # To log activation histograms to the forward pass, we register
            # a hook on forward to capture the output tensors.
            # This uses a closure on self._log_histograms_this_batch to
            # determine whether to send the activations to tensorboard,
            # since we don't want them on every call.
            for _, module in self._model.named_modules():
                if not getattr(module, u'should_log_activations', False):
                    # skip it
                    continue

                def hook(module_, inputs, outputs):
                    # pylint: disable=unused-argument,cell-var-from-loop
                    log_prefix = u'activation_histogram/{0}'.format(module_.__class__)
                    if self._log_histograms_this_batch:
                        if isinstance(outputs, torch.Tensor):
                            log_name = log_prefix
                            self._tensorboard.add_train_histogram(log_name,
                                                                  outputs,
                                                                  self._batch_num_total)
                        elif isinstance(outputs, (list, tuple)):
                            for i, output in enumerate(outputs):
                                log_name = u"{0}_{1}".format(log_prefix, i)
                                self._tensorboard.add_train_histogram(log_name,
                                                                      output,
                                                                      self._batch_num_total)
                        elif isinstance(outputs, dict):
                            for k, tensor in list(outputs.items()):
                                log_name = u"{0}_{1}".format(log_prefix, k)
                                self._tensorboard.add_train_histogram(log_name,
                                                                      tensor,
                                                                      self._batch_num_total)
                        else:
                            # skip it
                            pass

                module.register_forward_hook(hook)

    def _rescale_gradients(self)                   :
        u"""
        Performs gradient rescaling. Is a no-op if gradient rescaling is not enabled.
        """
        if self._grad_norm:
            parameters_to_clip = [p for p in self._model.parameters()
                                  if p.grad is not None]
            return sparse_clip_norm(parameters_to_clip, self._grad_norm)
        return None

    def _data_parallel(self, batch):
        u"""
        Do the forward pass using multiple GPUs.  This is a simplification
        of torch.nn.parallel.data_parallel to support the allennlp model
        interface.
        """
        inputs, module_kwargs = scatter_kwargs((), batch, self._cuda_devices, 0)
        used_device_ids = self._cuda_devices[:len(inputs)]
        replicas = replicate(self._model, used_device_ids)
        outputs = parallel_apply(replicas, inputs, module_kwargs, used_device_ids)

        # Only the 'loss' is needed.
        # a (num_gpu, ) tensor with loss on each GPU
        losses = gather([output[u'loss'].unsqueeze(0) for output in outputs], used_device_ids[0], 0)
        return {u'loss': losses.mean()}

    def _batch_loss(self, batch              , for_training      )                :
        u"""
        Does a forward pass on the given batch and returns the ``loss`` value in the result.
        If ``for_training`` is `True` also applies regularization penalty.
        """
        if self._multiple_gpu:
            output_dict = self._data_parallel(batch)
        else:
            output_dict = self._model(**batch)

        try:
            loss = output_dict[u"loss"]
            if for_training:
                loss += self._model.get_regularization_penalty()
        except KeyError:
            if for_training:
                raise RuntimeError(u"The model you are trying to optimize does not contain a"
                                   u" 'loss' key in the output of model.forward(inputs).")
            loss = None

        return loss

    def _get_metrics(self, total_loss       , num_batches     , reset       = False)                    :
        u"""
        Gets the metrics but sets ``"loss"`` to
        the total loss divided by the ``num_batches`` so that
        the ``"loss"`` metric is "average loss per batch".
        """
        metrics = self._model.get_metrics(reset=reset)
        metrics[u"loss"] = float(total_loss / num_batches) if num_batches > 0 else 0.0
        return metrics

    def _train_epoch(self, epoch     )                    :
        u"""
        Trains one epoch and returns metrics.
        """
        logger.info(u"Epoch %d/%d", epoch, self._num_epochs - 1)
        logger.info("Peak CPU memory usage MB: {peak_memory_mb()}")
        for gpu, memory in list(gpu_memory_mb().items()):
            logger.info("GPU {gpu} memory usage MB: {memory}")

        train_loss = 0.0
        # Set the model to "train" mode.
        self._model.train()

        # Get tqdm for the training batches
        train_generator = self._iterator(self._train_data,
                                         num_epochs=1,
                                         shuffle=self._shuffle,
                                         cuda_device=self._iterator_device)
        num_training_batches = self._iterator.get_num_batches(self._train_data)
        self._last_log = time.time()
        last_save_time = time.time()

        batches_this_epoch = 0
        if self._batch_num_total is None:
            self._batch_num_total = 0

        if self._histogram_interval is not None:
            histogram_parameters = set(self._model.get_parameters_for_histogram_tensorboard_logging())

        logger.info(u"Training")
        train_generator_tqdm = Tqdm.tqdm(train_generator,
                                         total=num_training_batches)
        for batch in train_generator_tqdm:
            batches_this_epoch += 1
            self._batch_num_total += 1
            batch_num_total = self._batch_num_total

            self._log_histograms_this_batch = self._histogram_interval is not None and (
                    batch_num_total % self._histogram_interval == 0)

            self._optimizer.zero_grad()

            loss = self._batch_loss(batch, for_training=True)
            loss.backward()

            train_loss += loss.item()

            batch_grad_norm = self._rescale_gradients()

            # This does nothing if batch_num_total is None or you are using an
            # LRScheduler which doesn't update per batch.
            if self._learning_rate_scheduler:
                self._learning_rate_scheduler.step_batch(batch_num_total)

            if self._log_histograms_this_batch:
                # get the magnitude of parameter updates for logging
                # We need a copy of current parameters to compute magnitude of updates,
                # and copy them to CPU so large models won't go OOM on the GPU.
                param_updates = dict((name, param.detach().cpu().clone())
                                 for name, param in self._model.named_parameters())
                self._optimizer.step()
                for name, param in self._model.named_parameters():
                    param_updates[name].sub_(param.detach().cpu())
                    update_norm = torch.norm(param_updates[name].view(-1, ))
                    param_norm = torch.norm(param.view(-1, )).cpu()
                    self._tensorboard.add_train_scalar(u"gradient_update/" + name,
                                                       update_norm / (param_norm + 1e-7),
                                                       batch_num_total)
            else:
                self._optimizer.step()

            # Update the description with the latest metrics
            metrics = self._get_metrics(train_loss, batches_this_epoch)
            description = self._description_from_metrics(metrics)

            train_generator_tqdm.set_description(description, refresh=False)

            # Log parameter values to Tensorboard
            if batch_num_total % self._summary_interval == 0:
                self._parameter_and_gradient_statistics_to_tensorboard(batch_num_total, batch_grad_norm)
                self._tensorboard.add_train_scalar(u"loss/loss_train", metrics[u"loss"], batch_num_total)
                self._metrics_to_tensorboard(batch_num_total,
                                             dict((u"epoch_metrics/" + k, v) for k, v in list(metrics.items())))

            if self._log_histograms_this_batch:
                self._histograms_to_tensorboard(batch_num_total, histogram_parameters)

            # Save model if needed.
            if self._model_save_interval is not None and (
                    time.time() - last_save_time > self._model_save_interval
            ):
                last_save_time = time.time()
                self._save_checkpoint(
                        u'{0}.{1}'.format(epoch, time_to_str(int(last_save_time))), [], is_best=False
                )

        return self._get_metrics(train_loss, batches_this_epoch, reset=True)

    def _should_stop_early(self, metric_history             )        :
        u"""
        uses patience and the validation metric to determine if training should stop early
        """
        if self._patience and self._patience < len(metric_history):
            # Pylint can't figure out that in this branch `self._patience` is an int.
            # pylint: disable=invalid-unary-operand-type

            # Is the best score in the past N epochs worse than or equal the best score overall?
            if self._validation_metric_decreases:
                return min(metric_history[-self._patience:]) >= min(metric_history[:-self._patience])
            else:
                return max(metric_history[-self._patience:]) <= max(metric_history[:-self._patience])

        return False

    def _parameter_and_gradient_statistics_to_tensorboard(self, # pylint: disable=invalid-name
                                                          epoch     ,
                                                          batch_grad_norm       )        :
        u"""
        Send the mean and std of all parameters and gradients to tensorboard, as well
        as logging the average gradient norm.
        """
        # Log parameter values to Tensorboard
        for name, param in self._model.named_parameters():
            self._tensorboard.add_train_scalar(u"parameter_mean/" + name,
                                               param.data.mean(),
                                               epoch)
            self._tensorboard.add_train_scalar(u"parameter_std/" + name, param.data.std(), epoch)
            if param.grad is not None:
                if is_sparse(param.grad):
                    # pylint: disable=protected-access
                    grad_data = param.grad.data._values()
                else:
                    grad_data = param.grad.data

                # skip empty gradients
                if torch.prod(torch.tensor(grad_data.shape)).item() > 0: # pylint: disable=not-callable
                    self._tensorboard.add_train_scalar(u"gradient_mean/" + name,
                                                       grad_data.mean(),
                                                       epoch)
                    self._tensorboard.add_train_scalar(u"gradient_std/" + name,
                                                       grad_data.std(),
                                                       epoch)
                else:
                    # no gradient for a parameter with sparse gradients
                    logger.info(u"No gradient for %s, skipping tensorboard logging.", name)
        # norm of gradients
        if batch_grad_norm is not None:
            self._tensorboard.add_train_scalar(u"gradient_norm",
                                               batch_grad_norm,
                                               epoch)

    def _histograms_to_tensorboard(self, epoch     , histogram_parameters          )        :
        u"""
        Send histograms of parameters to tensorboard.
        """
        for name, param in self._model.named_parameters():
            if name in histogram_parameters:
                self._tensorboard.add_train_histogram(u"parameter_histogram/" + name,
                                                      param,
                                                      epoch)

    def _metrics_to_tensorboard(self,
                                epoch     ,
                                train_metrics      ,
                                val_metrics       = None)        :
        u"""
        Sends all of the train metrics (and validation metrics, if provided) to tensorboard.
        """
        metric_names = set(train_metrics.keys())
        if val_metrics is not None:
            metric_names.update(list(val_metrics.keys()))
        val_metrics = val_metrics or {}

        for name in metric_names:
            train_metric = train_metrics.get(name)
            if train_metric is not None:
                self._tensorboard.add_train_scalar(name, train_metric, epoch)
            val_metric = val_metrics.get(name)
            if val_metric is not None:
                self._tensorboard.add_validation_scalar(name, val_metric, epoch)

    def _metrics_to_console(self,  # pylint: disable=no-self-use
                            train_metrics      ,
                            val_metrics       = None)        :
        u"""
        Logs all of the train metrics (and validation metrics, if provided) to the console.
        """
        val_metrics = val_metrics or {}
        dual_message_template = u"%s |  %8.3f  |  %8.3f"
        no_val_message_template = u"%s |  %8.3f  |  %8s"
        no_train_message_template = u"%s |  %8s  |  %8.3f"
        header_template = u"%s |  %-10s"

        metric_names = set(train_metrics.keys())
        if val_metrics:
            metric_names.update(list(val_metrics.keys()))

        name_length = max([len(x) for x in metric_names])

        logger.info(header_template, u"Training".rjust(name_length + 13), u"Validation")
        for name in metric_names:
            train_metric = train_metrics.get(name)
            val_metric = val_metrics.get(name)

            if val_metric is not None and train_metric is not None:
                logger.info(dual_message_template, name.ljust(name_length), train_metric, val_metric)
            elif val_metric is not None:
                logger.info(no_train_message_template, name.ljust(name_length), u"N/A", val_metric)
            elif train_metric is not None:
                logger.info(no_val_message_template, name.ljust(name_length), train_metric, u"N/A")

    def _validation_loss(self)                     :
        u"""
        Computes the validation loss. Returns it and the number of batches.
        """
        logger.info(u"Validating")

        self._model.eval()

        if self._validation_iterator is not None:
            val_iterator = self._validation_iterator
        else:
            val_iterator = self._iterator

        val_generator = val_iterator(self._validation_data,
                                     num_epochs=1,
                                     shuffle=False,
                                     cuda_device=self._iterator_device)
        num_validation_batches = val_iterator.get_num_batches(self._validation_data)
        val_generator_tqdm = Tqdm.tqdm(val_generator,
                                       total=num_validation_batches)
        batches_this_epoch = 0
        val_loss = 0
        for batch in val_generator_tqdm:

            loss = self._batch_loss(batch, for_training=False)
            if loss is not None:
                # You shouldn't necessarily have to compute a loss for validation, so we allow for
                # `loss` to be None.  We need to be careful, though - `batches_this_epoch` is
                # currently only used as the divisor for the loss function, so we can safely only
                # count those batches for which we actually have a loss.  If this variable ever
                # gets used for something else, we might need to change things around a bit.
                batches_this_epoch += 1
                val_loss += loss.detach().cpu().numpy()

            # Update the description with the latest metrics
            val_metrics = self._get_metrics(val_loss, batches_this_epoch)
            description = self._description_from_metrics(val_metrics)
            val_generator_tqdm.set_description(description, refresh=False)

        return val_loss, batches_this_epoch

    def train(self)                  :
        u"""
        Trains the supplied model with the supplied parameters.
        """
        try:
            epoch_counter, validation_metric_per_epoch = self._restore_checkpoint()
        except RuntimeError:
            traceback.print_exc()
            raise ConfigurationError(u"Could not recover training from the checkpoint.  Did you mean to output to "
                                     u"a different serialization directory or delete the existing serialization "
                                     u"directory?")

        self._enable_gradient_clipping()
        self._enable_activation_logging()

        logger.info(u"Beginning training.")

        train_metrics                   = {}
        val_metrics                   = {}
        best_epoch_val_metrics                   = {}
        epochs_trained = 0
        training_start_time = time.time()
        for epoch in range(epoch_counter, self._num_epochs):
            epoch_start_time = time.time()
            train_metrics = self._train_epoch(epoch)

            if self._validation_data is not None:
                with torch.no_grad():
                    # We have a validation set, so compute all the metrics on it.
                    val_loss, num_batches = self._validation_loss()
                    val_metrics = self._get_metrics(val_loss, num_batches, reset=True)

                    # Check validation metric for early stopping
                    this_epoch_val_metric = val_metrics[self._validation_metric]

                    # Check validation metric to see if it's the best so far
                    is_best_so_far = self._is_best_so_far(this_epoch_val_metric, validation_metric_per_epoch)
                    if is_best_so_far:
                        best_epoch_val_metrics = val_metrics.copy()
                    validation_metric_per_epoch.append(this_epoch_val_metric)
                    if self._should_stop_early(validation_metric_per_epoch):
                        logger.info(u"Ran out of patience.  Stopping training.")
                        break

            else:
                # No validation set, so just assume it's the best so far.
                is_best_so_far = True
                val_metrics = {}
                best_epoch_val_metrics = {}
                this_epoch_val_metric = None

            self._save_checkpoint(epoch, validation_metric_per_epoch, is_best=is_best_so_far)
            self._metrics_to_tensorboard(epoch, train_metrics, val_metrics=val_metrics)
            self._metrics_to_console(train_metrics, val_metrics)

            if self._learning_rate_scheduler:
                # The LRScheduler API is agnostic to whether your schedule requires a validation metric -
                # if it doesn't, the validation metric passed here is ignored.
                self._learning_rate_scheduler.step(this_epoch_val_metric, epoch)

            epoch_elapsed_time = time.time() - epoch_start_time
            logger.info(u"Epoch duration: %s", time.strftime(u"%H:%M:%S", time.gmtime(epoch_elapsed_time)))

            if epoch < self._num_epochs - 1:
                training_elapsed_time = time.time() - training_start_time
                estimated_time_remaining = training_elapsed_time *\
                    ((self._num_epochs - epoch_counter) / float(epoch - epoch_counter + 1) - 1)
                formatted_time = unicode(datetime.timedelta(seconds=int(estimated_time_remaining)))
                logger.info(u"Estimated training time remaining: %s", formatted_time)

            epochs_trained += 1

        training_elapsed_time = time.time() - training_start_time
        metrics = {
                u"training_duration": time.strftime(u"%H:%M:%S", time.gmtime(training_elapsed_time)),
                u"training_start_epoch": epoch_counter,
                u"training_epochs": epochs_trained
        }
        for key, value in list(train_metrics.items()):
            metrics[u"training_" + key] = value
        for key, value in list(val_metrics.items()):
            metrics[u"validation_" + key] = value

        if validation_metric_per_epoch:
            # We may not have had validation data, so we need to hide this behind an if.
            if self._validation_metric_decreases:
                best_validation_metric = min(validation_metric_per_epoch)
            else:
                best_validation_metric = max(validation_metric_per_epoch)
            metrics.update(dict(("best_validation_{k}", v) for k, v in list(best_epoch_val_metrics.items())))
            metrics[u'best_epoch'] = [i for i, value in enumerate(validation_metric_per_epoch)
                                     if value == best_validation_metric][-1]
        return metrics

    def _is_best_so_far(self,
                        this_epoch_val_metric       ,
                        validation_metric_per_epoch             ):
        if not validation_metric_per_epoch:
            return True
        elif self._validation_metric_decreases:
            return this_epoch_val_metric < min(validation_metric_per_epoch)
        else:
            return this_epoch_val_metric > max(validation_metric_per_epoch)

    def _description_from_metrics(self, metrics                  )       :
        if (not self._warned_tqdm_ignores_underscores and
                    any(metric_name.startswith(u"_") for metric_name in metrics)):
            logger.warning(u"Metrics with names beginning with \"_\" will "
                           u"not be logged to the tqdm progress bar.")
            self._warned_tqdm_ignores_underscores = True
        return u', '.join([u"%s: %.4f" % (name, value) for name, value in
                          list(metrics.items()) if not name.startswith(u"_")]) + u" ||"

    def _save_checkpoint(self,
                         epoch                 ,
                         val_metric_per_epoch             ,
                         is_best                 = None)        :
        u"""
        Saves a checkpoint of the model to self._serialization_dir.
        Is a no-op if self._serialization_dir is None.

        Parameters
        ----------
        epoch : Union[int, str], required.
            The epoch of training.  If the checkpoint is saved in the middle
            of an epoch, the parameter is a string with the epoch and timestamp.
        is_best: bool, optional (default = None)
            A flag which causes the model weights at the given epoch to
            be copied to a "best.th" file. The value of this flag should
            be based on some validation metric computed by your model.
        """
        if self._serialization_dir is not None:
            model_path = os.path.join(self._serialization_dir, u"model_state_epoch_{}.th".format(epoch))
            model_state = self._model.state_dict()
            torch.save(model_state, model_path)

            training_state = {u'epoch': epoch,
                              u'val_metric_per_epoch': val_metric_per_epoch,
                              u'optimizer': self._optimizer.state_dict(),
                              u'batch_num_total': self._batch_num_total}
            training_path = os.path.join(self._serialization_dir,
                                         u"training_state_epoch_{}.th".format(epoch))
            torch.save(training_state, training_path)
            if is_best:
                logger.info(u"Best validation performance so far. "
                            u"Copying weights to '%s/best.th'.", self._serialization_dir)
                shutil.copyfile(model_path, os.path.join(self._serialization_dir, u"best.th"))

            if self._num_serialized_models_to_keep and self._num_serialized_models_to_keep >= 0:
                self._serialized_paths.append([time.time(), model_path, training_path])
                if len(self._serialized_paths) > self._num_serialized_models_to_keep:
                    paths_to_remove = self._serialized_paths.pop(0)
                    # Check to see if we should keep this checkpoint, if it has been longer
                    # then self._keep_serialized_model_every_num_seconds since the last
                    # kept checkpoint.
                    remove_path = True
                    if self._keep_serialized_model_every_num_seconds is not None:
                        save_time = paths_to_remove[0]
                        time_since_checkpoint_kept = save_time - self._last_permanent_saved_checkpoint_time
                        if time_since_checkpoint_kept > self._keep_serialized_model_every_num_seconds:
                            # We want to keep this checkpoint.
                            remove_path = False
                            self._last_permanent_saved_checkpoint_time = save_time
                    if remove_path:
                        for fname in paths_to_remove[1:]:
                            os.remove(fname)

    def find_latest_checkpoint(self)                   :
        u"""
        Return the location of the latest model and training state files.
        If there isn't a valid checkpoint then return None.
        """
        have_checkpoint = (self._serialization_dir is not None and
                           any(u"model_state_epoch_" in x for x in os.listdir(self._serialization_dir)))

        if not have_checkpoint:
            return None

        serialization_files = os.listdir(self._serialization_dir)
        model_checkpoints = [x for x in serialization_files if u"model_state_epoch" in x]
        # Get the last checkpoint file.  Epochs are specified as either an
        # int (for end of epoch files) or with epoch and timestamp for
        # within epoch checkpoints, e.g. 5.2018-02-02-15-33-42
        found_epochs = [
                # pylint: disable=anomalous-backslash-in-string
                re.search(u"model_state_epoch_([0-9\.\-]+)\.th", x).group(1)
                for x in model_checkpoints
        ]
        int_epochs      = []
        for epoch in found_epochs:
            pieces = epoch.split(u'.')
            if len(pieces) == 1:
                # Just a single epoch without timestamp
                int_epochs.append([int(pieces[0]), 0])
            else:
                # has a timestamp
                int_epochs.append([int(pieces[0]), pieces[1]])
        last_epoch = sorted(int_epochs, reverse=True)[0]
        if last_epoch[1] == 0:
            epoch_to_load = unicode(last_epoch[0])
        else:
            epoch_to_load = u'{0}.{1}'.format(last_epoch[0], last_epoch[1])

        model_path = os.path.join(self._serialization_dir,
                                  u"model_state_epoch_{}.th".format(epoch_to_load))
        training_state_path = os.path.join(self._serialization_dir,
                                           u"training_state_epoch_{}.th".format(epoch_to_load))

        return (model_path, training_state_path)

    def _restore_checkpoint(self)                           :
        u"""
        Restores a model from a serialization_dir to the last saved checkpoint.
        This includes an epoch count and optimizer state, which is serialized separately
        from  model parameters. This function should only be used to continue training -
        if you wish to load a model for inference/load parts of a model into a new
        computation graph, you should use the native Pytorch functions:
        `` model.load_state_dict(torch.load("/path/to/model/weights.th"))``

        If ``self._serialization_dir`` does not exist or does not contain any checkpointed weights,
        this function will do nothing and return 0.

        Returns
        -------
        epoch: int
            The epoch at which to resume training, which should be one after the epoch
            in the saved training state.
        """
        latest_checkpoint = self.find_latest_checkpoint()

        if latest_checkpoint is None:
            # No checkpoint to restore, start at 0
            return 0, []

        model_path, training_state_path = latest_checkpoint

        # Load the parameters onto CPU, then transfer to GPU.
        # This avoids potential OOM on GPU for large models that
        # load parameters onto GPU then make a new GPU copy into the parameter
        # buffer. The GPU transfer happens implicitly in load_state_dict.
        model_state = torch.load(model_path, map_location=util.device_mapping(-1))
        training_state = torch.load(training_state_path, map_location=util.device_mapping(-1))
        self._model.load_state_dict(model_state)
        self._optimizer.load_state_dict(training_state[u"optimizer"])
        move_optimizer_to_cuda(self._optimizer)

        # We didn't used to save `validation_metric_per_epoch`, so we can't assume
        # that it's part of the trainer state. If it's not there, an empty list is all
        # we can do.
        if u"val_metric_per_epoch" not in training_state:
            logger.warning(u"trainer state `val_metric_per_epoch` not found, using empty list")
            val_metric_per_epoch              = []
        else:
            val_metric_per_epoch = training_state[u"val_metric_per_epoch"]

        if isinstance(training_state[u"epoch"], int):
            epoch_to_return = training_state[u"epoch"] + 1
        else:
            epoch_to_return = int(training_state[u"epoch"].split(u'.')[0]) + 1

        # For older checkpoints with batch_num_total missing, default to old behavior where
        # it is unchanged.
        batch_num_total = training_state.get(u'batch_num_total')
        if batch_num_total is not None:
            self._batch_num_total = batch_num_total

        return epoch_to_return, val_metric_per_epoch

    # Requires custom from_params.
    @classmethod
    def from_params(cls,
                    model       ,
                    serialization_dir     ,
                    iterator              ,
                    train_data                    ,
                    validation_data                              ,
                    params        ,
                    validation_iterator               = None)             :

        patience = params.pop_int(u"patience", None)
        validation_metric = params.pop(u"validation_metric", u"-loss")
        shuffle = params.pop_bool(u"shuffle", True)
        num_epochs = params.pop_int(u"num_epochs", 20)
        cuda_device = params.pop_int(u"cuda_device", -1)
        grad_norm = params.pop_float(u"grad_norm", None)
        grad_clipping = params.pop_float(u"grad_clipping", None)
        lr_scheduler_params = params.pop(u"learning_rate_scheduler", None)

        if cuda_device >= 0:
            model = model.cuda(cuda_device)
        parameters = [[n, p] for n, p in model.named_parameters() if p.requires_grad]
        optimizer = Optimizer.from_params(parameters, params.pop(u"optimizer"))

        if lr_scheduler_params:
            scheduler = LearningRateScheduler.from_params(optimizer, lr_scheduler_params)
        else:
            scheduler = None

        num_serialized_models_to_keep = params.pop_int(u"num_serialized_models_to_keep", 20)
        keep_serialized_model_every_num_seconds = params.pop_int(
                u"keep_serialized_model_every_num_seconds", None)
        model_save_interval = params.pop_float(u"model_save_interval", None)
        summary_interval = params.pop_int(u"summary_interval", 100)
        histogram_interval = params.pop_int(u"histogram_interval", None)

        params.assert_empty(cls.__name__)
        return Trainer(model, optimizer, iterator,
                       train_data, validation_data,
                       patience=patience,
                       validation_metric=validation_metric,
                       validation_iterator=validation_iterator,
                       shuffle=shuffle,
                       num_epochs=num_epochs,
                       serialization_dir=serialization_dir,
                       cuda_device=cuda_device,
                       grad_norm=grad_norm,
                       grad_clipping=grad_clipping,
                       learning_rate_scheduler=scheduler,
                       num_serialized_models_to_keep=num_serialized_models_to_keep,
                       keep_serialized_model_every_num_seconds=keep_serialized_model_every_num_seconds,
                       model_save_interval=model_save_interval,
                       summary_interval=summary_interval,
                       histogram_interval=histogram_interval)



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
import pytest

from allennlp.data.dataset_readers import QuoraParaphraseDatasetReader
from allennlp.common.util import ensure_list
from allennlp.common.testing import AllenNlpTestCase
try:
    from itertools import izip
except:
    izip = zip



class TestQuoraParaphraseReader(object):
    @pytest.mark.parametrize(u"lazy", (True, False))
    def test_read_from_file(self, lazy):
        reader = QuoraParaphraseDatasetReader(lazy=lazy)
        instances = reader.read(AllenNlpTestCase.FIXTURES_ROOT / u'data' / u'quora_paraphrase.tsv')
        instances = ensure_list(instances)

        instance1 = {u"premise": u"What should I do to avoid sleeping in class ?".split(),
                     u"hypothesis": u"How do I not sleep in a boring class ?".split(),
                     u"label": u"1"}

        instance2 = {u"premise": u"Do women support each other more than men do ?".split(),
                     u"hypothesis": u"Do women need more compliments than men ?".split(),
                     u"label": u"0"}

        instance3 = {u"premise": u"How can one root android devices ?".split(),
                     u"hypothesis": u"How do I root an Android device ?".split(),
                     u"label": u"1"}

        assert len(instances) == 3

        for instance, expected_instance in izip(instances, [instance1, instance2, instance3]):
            fields = instance.fields
            assert [t.text for t in fields[u"premise"].tokens] == expected_instance[u"premise"]
            assert [t.text for t in fields[u"hypothesis"].tokens] == expected_instance[u"hypothesis"]
            assert fields[u"label"].label == expected_instance[u"label"]



# pylint: disable=no-self-use,line-too-long


from __future__ import absolute_import
from __future__ import print_function
from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor

def demo_model(archive_file     , predictor_name     )             :
    archive = load_archive(archive_file)
    return Predictor.from_archive(archive, predictor_name)

# TODO(joelgrus): this is duplicated in the demo repo
# figure out where it really belongs
DEFAULT_MODELS = {
        u'machine-comprehension': (
                u'https://s3-us-west-2.amazonaws.com/allennlp/models/bidaf-model-2017.09.15-charpad.tar.gz',  # pylint: disable=line-too-long
                u'machine-comprehension'
        ),
        u'semantic-role-labeling': (
                u'https://s3-us-west-2.amazonaws.com/allennlp/models/srl-model-2018.05.25.tar.gz', # pylint: disable=line-too-long
                u'semantic-role-labeling'
        ),
        u'textual-entailment': (
                u'https://s3-us-west-2.amazonaws.com/allennlp/models/decomposable-attention-elmo-2018.02.19.tar.gz',  # pylint: disable=line-too-long
                u'textual-entailment'
        ),
        u'coreference-resolution': (
                u'https://s3-us-west-2.amazonaws.com/allennlp/models/coref-model-2018.02.05.tar.gz',  # pylint: disable=line-too-long
                u'coreference-resolution'
        ),
        u'named-entity-recognition': (
                u'https://s3-us-west-2.amazonaws.com/allennlp/models/ner-model-2018.04.30.tar.gz',  # pylint: disable=line-too-long
                u'sentence-tagger'
        ),
        u'constituency-parsing': (
                u'https://s3-us-west-2.amazonaws.com/allennlp/models/elmo-constituency-parser-2018.03.14.tar.gz',  # pylint: disable=line-too-long
                u'constituency-parser'
        ),
        u'dependency-parsing': (
                u'https://s3-us-west-2.amazonaws.com/allennlp/models/biaffine-dependency-parser-2018.08.01.tar.gz',  # pylint: disable=line-too-long
                u'biaffine-dependency-parser'
        )
}


class SniffTest(AllenNlpTestCase):

    def test_machine_comprehension(self):
        predictor = demo_model(*DEFAULT_MODELS[u'machine-comprehension'])

        passage = u"""The Matrix is a 1999 science fiction action film written and directed by The Wachowskis, starring Keanu Reeves, Laurence Fishburne, Carrie-Anne Moss, Hugo Weaving, and Joe Pantoliano. It depicts a dystopian future in which reality as perceived by most humans is actually a simulated reality called "the Matrix", created by sentient machines to subdue the human population, while their bodies' heat and electrical activity are used as an energy source. Computer programmer Neo" learns this truth and is drawn into a rebellion against the machines, which involves other people who have been freed from the "dream world". """  # pylint: disable=line-too-long
        question = u"Who stars in The Matrix?"

        result = predictor.predict_json({u"passage": passage, u"question": question})

        correct = u"Keanu Reeves, Laurence Fishburne, Carrie-Anne Moss, Hugo Weaving, and Joe Pantoliano"

        assert correct == result[u"best_span_str"]

    def test_semantic_role_labeling(self):
        predictor = demo_model(*DEFAULT_MODELS[u'semantic-role-labeling'])

        sentence = u"If you liked the music we were playing last night, you will absolutely love what we're playing tomorrow!"

        result = predictor.predict_json({u"sentence": sentence})

        assert result[u"words"] == [
                u"If", u"you", u"liked", u"the", u"music", u"we", u"were", u"playing", u"last", u"night", u",",
                u"you", u"will", u"absolutely", u"love", u"what", u"we", u"'re", u"playing", u"tomorrow", u"!"
        ]

        assert result[u"verbs"] == [
                {u"verb": u"liked",
                 u"description": u"If [ARG0: you] [V: liked] [ARG1: the music we were playing last night] , you will absolutely love what we 're playing tomorrow !",
                 u"tags": [u"O", u"B-ARG0", u"B-V", u"B-ARG1", u"I-ARG1", u"I-ARG1", u"I-ARG1", u"I-ARG1", u"I-ARG1", u"I-ARG1", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O"]},
                {u"verb": u"were",
                 u"description": u"If you liked the music we [V: were] playing last night , you will absolutely love what we 're playing tomorrow !",
                 u"tags": [u"O", u"O", u"O", u"O", u"O", u"O", u"B-V", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O"]},
                {u"verb": u"playing",
                 u"description": u"If you liked [ARG1: the music] [ARG0: we] were [V: playing] [ARGM-TMP: last] night , you will absolutely love what we 're playing tomorrow !",
                 u"tags": [u"O", u"O", u"O", u"B-ARG1", u"I-ARG1", u"B-ARG0", u"O", u"B-V", u"B-ARGM-TMP", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O"]},
                {u"verb": u"will",
                 u"description": u"If you liked the music we were playing last night , you [V: will] absolutely love what we 're playing tomorrow !",
                 u"tags": [u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"B-V", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O"]},
                {u"verb": u"love",
                 u"description": u"[ARGM-ADV: If you liked the music we were playing last night] , [ARG0: you] [ARGM-MOD: will] [ARGM-ADV: absolutely] [V: love] [ARG1: what we 're playing tomorrow] !",
                 u"tags": [u"B-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"I-ARGM-ADV", u"O", u"B-ARG0", u"B-ARGM-MOD", u"B-ARGM-ADV", u"B-V", u"B-ARG1", u"I-ARG1", u"I-ARG1", u"I-ARG1", u"I-ARG1", u"O"]},
                {u"verb": u"'re",
                 u"description": u"If you liked the music we were playing last night , you will absolutely love what we [V: 're] playing tomorrow !",
                 u"tags": [u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"B-V", u"O", u"O", u"O"]},
                {u"verb": u"playing",
                 u"description": u"If you liked the music we were playing last night , you will absolutely love [ARG1: what] [ARG0: we] 're [V: playing] [ARGM-TMP: tomorrow] !",
                 u"tags": [u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"O", u"B-ARG1", u"B-ARG0", u"O", u"B-V", u"B-ARGM-TMP", u"O"]}
        ]

    def test_textual_entailment(self):
        predictor = demo_model(*DEFAULT_MODELS[u'textual-entailment'])

        result = predictor.predict_json({
                u"premise": u"An interplanetary spacecraft is in orbit around a gas giant's icy moon.",
                u"hypothesis": u"The spacecraft has the ability to travel between planets."
        })

        assert result[u"label_probs"][0] > 0.7  # entailment

        result = predictor.predict_json({
                u"premise": u"Two women are wandering along the shore drinking iced tea.",
                u"hypothesis": u"Two women are sitting on a blanket near some rocks talking about politics."
        })

        assert result[u"label_probs"][1] > 0.8  # contradiction

        result = predictor.predict_json({
                u"premise": u"A large, gray elephant walked beside a herd of zebras.",
                u"hypothesis": u"The elephant was lost."
        })

        assert result[u"label_probs"][2] > 0.6  # neutral

    def test_coreference_resolution(self):
        predictor = demo_model(*DEFAULT_MODELS[u'coreference-resolution'])

        document = u"We 're not going to skimp on quality , but we are very focused to make next year . The only problem is that some of the fabrics are wearing out - since I was a newbie I skimped on some of the fabric and the poor quality ones are developing holes ."

        result = predictor.predict_json({u"document": document})
        print(result)
        assert result[u'clusters'] == [[[0, 0], [10, 10]],
                                      [[33, 33], [37, 37]],
                                      [[26, 27], [42, 43]]]
        assert result[u"document"] == [u'We', u"'re", u'not', u'going', u'to', u'skimp', u'on', u'quality', u',', u'but', u'we', u'are',
                                      u'very', u'focused', u'to', u'make', u'next', u'year', u'.', u'The', u'only', u'problem', u'is',
                                      u'that', u'some', u'of', u'the', u'fabrics', u'are', u'wearing', u'out', u'-', u'since', u'I', u'was',
                                      u'a', u'newbie', u'I', u'skimped', u'on', u'some', u'of', u'the', u'fabric', u'and', u'the', u'poor',
                                      u'quality', u'ones', u'are', u'developing', u'holes', u'.']

    def test_ner(self):
        predictor = demo_model(*DEFAULT_MODELS[u'named-entity-recognition'])

        sentence = u"""Michael Jordan is a professor at Berkeley."""

        result = predictor.predict_json({u"sentence": sentence})

        assert result[u"words"] == [u"Michael", u"Jordan", u"is", u"a", u"professor", u"at", u"Berkeley", u"."]
        assert result[u"tags"] == [u"B-PER", u"L-PER", u"O", u"O", u"O", u"O", u"U-LOC", u"O"]

    def test_constituency_parsing(self):
        predictor = demo_model(*DEFAULT_MODELS[u'constituency-parsing'])

        sentence = u"""Pierre Vinken died aged 81; immortalised aged 61."""

        result = predictor.predict_json({u"sentence": sentence})

        assert result[u"tokens"] == [u"Pierre", u"Vinken", u"died", u"aged", u"81", u";", u"immortalised", u"aged", u"61", u"."]
        assert result[u"trees"] == u"(S (NP (NNP Pierre) (NNP Vinken)) (VP (VP (VBD died) (NP (JJ aged) (CD 81))) (, ;) (VP (VBD immortalised) (S (ADJP (VBN aged) (NP (CD 61)))))) (. .))"

    def test_dependency_parsing(self):
        predictor = demo_model(*DEFAULT_MODELS[u'dependency-parsing'])
        sentence = u"""He ate spaghetti with chopsticks."""
        result = predictor.predict_json({u"sentence": sentence})
        # Note that this tree is incorrect. We are checking here that the decoded
        # tree is _actually a tree_ - in greedy decoding versions of the dependency
        # parser, this sentence has multiple heads. This test shouldn't really live here,
        # but it's very difficult to re-create a concrete example of this behaviour without
        # a trained dependency parser.
        assert result[u'words'] == [u'He', u'ate', u'spaghetti', u'with', u'chopsticks', u'.']
        assert result[u'pos'] == [u'PRP', u'VBD', u'NNS', u'IN', u'NNS', u'.']
        assert result[u'predicted_dependencies'] == [u'nsubj', u'root', u'acomp', u'prep', u'pobj', u'punct']
        assert result[u'predicted_heads'] == [3, 0, 2, 3, 4, 2]




from __future__ import absolute_import
from collections import defaultdict
#typing

from allennlp.common.registrable import FromParams
from allennlp.nn.decoding.decoder_step import DecoderStep
from allennlp.nn.decoding.decoder_state import DecoderState


class BeamSearch(FromParams):
    u"""
    This class implements beam search over transition sequences given an initial ``DecoderState``
    and a ``DecoderStep``, returning the highest scoring final states found by the beam (the states
    will keep track of the transition sequence themselves).

    The initial ``DecoderState`` is assumed to be `batched`.  The value we return from the search
    is a dictionary from batch indices to ranked finished states.

    IMPORTANT: We assume that the ``DecoderStep`` that you are using returns possible next states
    in sorted order, so we do not do an additional sort inside of ``BeamSearch.search()``.  If
    you're implementing your own ``DecoderStep``, you must ensure that you've sorted the states
    that you return.
    """
    def __init__(self, beam_size     )        :
        self._beam_size = beam_size

    def search(self,
               num_steps     ,
               initial_state              ,
               decoder_step             ,
               keep_final_unfinished_states       = True)                                 :
        u"""
        Parameters
        ----------
        num_steps : ``int``
            How many steps should we take in our search?  This is an upper bound, as it's possible
            for the search to run out of valid actions before hitting this number, or for all
            states on the beam to finish.
        initial_state : ``DecoderState``
            The starting state of our search.  This is assumed to be `batched`, and our beam search
            is batch-aware - we'll keep ``beam_size`` states around for each instance in the batch.
        decoder_step : ``DecoderStep``
            The ``DecoderStep`` object that defines and scores transitions from one state to the
            next.
        keep_final_unfinished_states : ``bool``, optional (default=True)
            If we run out of steps before a state is "finished", should we return that state in our
            search results?

        Returns
        -------
        best_states : ``Dict[int, List[DecoderState]]``
            This is a mapping from batch index to the top states for that instance.
        """
        finished_states                                = defaultdict(list)
        states = [initial_state]
        step_num = 1
        while states and step_num <= num_steps:
            next_states                                = defaultdict(list)
            grouped_state = states[0].combine_states(states)
            for next_state in decoder_step.take_step(grouped_state, max_actions=self._beam_size):
                # NOTE: we're doing state.batch_indices[0] here (and similar things below),
                # hard-coding a group size of 1.  But, our use of `next_state.is_finished()`
                # already checks for that, as it crashes if the group size is not 1.
                batch_index = next_state.batch_indices[0]
                if next_state.is_finished():
                    finished_states[batch_index].append(next_state)
                else:
                    if step_num == num_steps and keep_final_unfinished_states:
                        finished_states[batch_index].append(next_state)
                    next_states[batch_index].append(next_state)
            states = []
            for batch_index, batch_states in list(next_states.items()):
                # The states from the generator are already sorted, so we can just take the first
                # ones here, without an additional sort.
                states.extend(batch_states[:self._beam_size])
            step_num += 1
        best_states                                = {}
        for batch_index, batch_states in list(finished_states.items()):
            # The time this sort takes is pretty negligible, no particular need to optimize this
            # yet.  Maybe with a larger beam size...
            finished_to_sort = [(-state.score[0].item(), state) for state in batch_states]
            finished_to_sort.sort(key=lambda x: x[0])
            best_states[batch_index] = [state[1] for state in finished_to_sort[:self._beam_size]]
        return best_states




from __future__ import absolute_import
#typing

#overrides
from nltk import Tree
from spacy.lang.en.tag_map import TAG_MAP

from allennlp.common.util import JsonDict, sanitize
from allennlp.data import DatasetReader, Instance
from allennlp.models import Model
from allennlp.predictors.predictor import Predictor
from allennlp.data.tokenizers.word_splitter import SpacyWordSplitter


# Make the links to POS tag nodes render as "pos",
# to distinguish them from constituency tags. The
# actual tag is still visible within the node.
LINK_TO_LABEL = dict((x, u"pos") for x in TAG_MAP)

# POS tags have a unified colour.
NODE_TYPE_TO_STYLE = dict((x, [u"color0"]) for x in TAG_MAP)

# Verb and Noun phrases get their own colour.
NODE_TYPE_TO_STYLE[u"NP"] = [u"color1"]
NODE_TYPE_TO_STYLE[u"NX"] = [u"color1"]
NODE_TYPE_TO_STYLE[u"QP"] = [u"color1"]
NODE_TYPE_TO_STYLE[u"NAC"] = [u"color1"]
NODE_TYPE_TO_STYLE[u"VP"] = [u"color2"]

# Clause level fragments
NODE_TYPE_TO_STYLE[u"S"] = [u"color3"]
NODE_TYPE_TO_STYLE[u"SQ"] = [u"color3"]
NODE_TYPE_TO_STYLE[u"SBAR"] = [u"color3"]
NODE_TYPE_TO_STYLE[u"SBARQ"] = [u"color3"]
NODE_TYPE_TO_STYLE[u"SINQ"] = [u"color3"]
NODE_TYPE_TO_STYLE[u"FRAG"] = [u"color3"]
NODE_TYPE_TO_STYLE[u"X"] = [u"color3"]

# Wh-phrases.
NODE_TYPE_TO_STYLE[u"WHADVP"] = [u"color4"]
NODE_TYPE_TO_STYLE[u"WHADJP"] = [u"color4"]
NODE_TYPE_TO_STYLE[u"WHNP"] = [u"color4"]
NODE_TYPE_TO_STYLE[u"WHPP"] = [u"color4"]

# Prepositional Phrases get their own colour because
# they are linguistically interesting.
NODE_TYPE_TO_STYLE[u"PP"] = [u"color6"]

# Everything else.
NODE_TYPE_TO_STYLE[u"ADJP"] = [u"color5"]
NODE_TYPE_TO_STYLE[u"ADVP"] = [u"color5"]
NODE_TYPE_TO_STYLE[u"CONJP"] = [u"color5"]
NODE_TYPE_TO_STYLE[u"INTJ"] = [u"color5"]
NODE_TYPE_TO_STYLE[u"LST"] = [u"color5", u"seq"]
NODE_TYPE_TO_STYLE[u"PRN"] = [u"color5"]
NODE_TYPE_TO_STYLE[u"PRT"] = [u"color5"]
NODE_TYPE_TO_STYLE[u"RRC"] = [u"color5"]
NODE_TYPE_TO_STYLE[u"UCP"] = [u"color5"]


class ConstituencyParserPredictor(Predictor):
    u"""
    Predictor for the :class:`~allennlp.models.SpanConstituencyParser` model.
    """
    def __init__(self, model       , dataset_reader               )        :
        super(ConstituencyParserPredictor, self).__init__(model, dataset_reader)
        self._tokenizer = SpacyWordSplitter(language=u'en_core_web_sm', pos_tags=True)

    def predict(self, sentence     )            :
        u"""
        Predict a constituency parse for the given sentence.
        Parameters
        ----------
        sentence The sentence to parse.

        Returns
        -------
        A dictionary representation of the constituency tree.
        """
        return self.predict_json({u"sentence" : sentence})

    #overrides
    def _json_to_instance(self, json_dict          )            :
        u"""
        Expects JSON that looks like ``{"sentence": "..."}``.
        """
        spacy_tokens = self._tokenizer.split_words(json_dict[u"sentence"])
        sentence_text = [token.text for token in spacy_tokens]
        pos_tags = [token.tag_ for token in spacy_tokens]
        return self._dataset_reader.text_to_instance(sentence_text, pos_tags)

    #overrides
    def predict_instance(self, instance          )            :
        outputs = self._model.forward_on_instance(instance)

        # format the NLTK tree as a string on a single line.
        tree = outputs.pop(u"trees")
        outputs[u"hierplane_tree"] = self._build_hierplane_tree(tree, 0, is_root=True)
        outputs[u"trees"] = tree.pformat(margin=1000000)
        return sanitize(outputs)

    #overrides
    def predict_batch_instance(self, instances                )                  :
        outputs = self._model.forward_on_instances(instances)
        for output in outputs:
            # format the NLTK tree as a string on a single line.
            tree = output.pop(u"trees")
            output[u"hierplane_tree"] = self._build_hierplane_tree(tree, 0, is_root=True)
            output[u"trees"] = tree.pformat(margin=1000000)
        return sanitize(outputs)


    def _build_hierplane_tree(self, tree      , index     , is_root      )            :
        u"""
        Recursively builds a JSON dictionary from an NLTK ``Tree`` suitable for
        rendering trees using the `Hierplane library<https://allenai.github.io/hierplane/>`.

        Parameters
        ----------
        tree : ``Tree``, required.
            The tree to convert into Hierplane JSON.
        index : int, required.
            The character index into the tree, used for creating spans.
        is_root : bool
            An indicator which allows us to add the outer Hierplane JSON which
            is required for rendering.

        Returns
        -------
        A JSON dictionary render-able by Hierplane for the given tree.
        """
        children = []
        for child in tree:
            if isinstance(child, Tree):
                # If the child is a tree, it has children,
                # as NLTK leaves are just strings.
                children.append(self._build_hierplane_tree(child, index, is_root=False))
            else:
                # We're at a leaf, so add the length of
                # the word to the character index.
                index += len(child)

        label = tree.label()
        span = u" ".join(tree.leaves())
        hierplane_node = {
                u"word": span,
                u"nodeType": label,
                u"attributes": [label],
                u"link": label
        }
        if children:
            hierplane_node[u"children"] = children
        # TODO(Mark): Figure out how to span highlighting to the leaves.
        if is_root:
            hierplane_node = {
                    u"linkNameToLabel": LINK_TO_LABEL,
                    u"nodeTypeToStyle": NODE_TYPE_TO_STYLE,
                    u"text": span,
                    u"root": hierplane_node
            }
        return hierplane_node

ConstituencyParserPredictor = Predictor.register(u'constituency-parser')(ConstituencyParserPredictor)



#!/usr/bin/env python
#
# See the accompanying LICENSE file.
#
# This module lets you automatically trace SQL operations in a program
# using APSW without having to modify the program in any way.

import time
import sys
import weakref

class APSWTracer(object):

    def __init__(self, options):
        if sys.version_info<(3,):
            self.u=eval("u''")
            import thread
            self.threadid=thread.get_ident
            self.stringtypes=(unicode,str)
            self.numtypes=(int, long, float)
            self.binarytypes=(buffer,)
        else:
            self.u=""
            import _thread
            self.threadid=_thread.get_ident
            self.stringtypes=(str,)
            self.numtypes=(int, float)
            self.binarytypes=(bytes,)
        self.options=options
        if options.output in ("-", "stdout"):
            self._writer=sys.stdout.write
        elif options.output=="stderr":
            self._writer=sys.stderr.write
        else:
            self._writer=open(options.output, "wt").write

        try:
            import apsw
            apsw.connection_hooks.append(self.connection_hook)
        except:
            sys.stderr.write(self.u+"Unable to import apsw\n")
            raise

        self.mapping_open_flags=apsw.mapping_open_flags
        self.zeroblob=apsw.zeroblob
        self.apswConnection=apsw.Connection

        self.newcursor={}
        self.threadsused={} # really want a set
        self.queries={}
        self.timings={}
        self.rowsreturned=0
        self.numcursors=0
        self.numconnections=0
        self.timestart=time.time()

    def writerpy2(self, s):
        # s should be a unicode string
        self._writer(s.encode("utf-8")+"\n")

    def writerpy3(self, s):
        self._writer(s+"\n")

    if sys.version_info<(3,):
        writer=writerpy2
    else:
        writer=writerpy3

    def format(self, obj):
        if isinstance(obj, dict):
            return self.formatdict(obj)
        if isinstance(obj, tuple):
            return self.formatseq(obj, '()')
        if isinstance(obj, list):
            return self.formatseq(obj, '[]')
        if isinstance(obj, self.stringtypes):
            return self.formatstring(obj)
        if obj is True:
            return "True"
        if obj is False:
            return "False"
        if obj is None:
            return "None"
        if isinstance(obj, self.numtypes):
            return repr(obj)
        if isinstance(obj, self.binarytypes):
            return self.formatbinary(obj)
        if isinstance(obj, self.zeroblob):
            return "zeroblob(%d)" % (obj.length(),)
        return repr(obj)

    def formatstring(self, obj, quote='"', checkmaxlen=True):
        obj=obj.replace("\n", "\\n").replace("\r", "\\r")
        if checkmaxlen and len(obj)>self.options.length:
            obj=obj[:self.options.length]+'..'
        return self.u+quote+obj+quote

    def formatdict(self, obj):
        items=list(obj.items())
        items.sort()
        op=[]
        for k,v in items:
            op.append(self.format(k)+": "+self.format(v))
        return self.u+"{"+", ".join(op)+"}"

    def formatseq(self, obj, paren):
        return self.u+paren[0]+", ".join([self.format(v) for v in obj])+paren[1]

    def formatbinarypy2(self, obj):
        if len(obj)<self.options.length:
            return "X'"+"".join(["%x" % ord(obj[i]) for i in range(len(obj))])+"'"
        return "(%d) X'"%(len(obj),)+"".join(["%x" % ord(obj[i]) for i in range(self.options.length)])+"..'"

    def formatbinarypy3(self, obj):
        if len(obj)<self.options.length:
            return "X'"+"".join(["%x" % obj[i] for i in range(len(obj))])+"'"
        return "(%d) X'"%(len(obj),)+"".join(["%x" % obj[i] for i in range(self.options.length)])+"..'"

    if sys.version_info<(3,):
        formatbinary=formatbinarypy2
    else:
        formatbinary=formatbinarypy3

    def sanitizesql(self, sql):
        sql=sql.strip("; \t\r\n")
        while sql.startswith("--"):
            sql=sql.split("\n", 1)[1]
            sql=sql.lstrip("; \t\r\n")
        return sql
            
    def profiler(self, sql, nanoseconds):
        sql=self.sanitizesql(sql)
        if sql not in self.timings:
            self.timings[sql]=[nanoseconds]
        else:
            self.timings[sql].append(nanoseconds)

    def cursorfinished(self, cursor):
        del self.newcursor[cursor]

    def exectracer(self, cursor, sql, bindings):
        tid=self.threadid()
        if tid not in self.threadsused:
            self.threadsused[tid]=True
        if self.options.report:
            fix=self.sanitizesql(sql)
            if fix not in self.queries:
                self.queries[fix]=1
            else:
                self.queries[fix]=self.queries[fix]+1
        if not isinstance(cursor, self.apswConnection):
            wr=weakref.ref(cursor, self.cursorfinished)
            if wr not in self.newcursor:
                self.newcursor[wr]=True
                self.numcursors+=1
                if self.options.sql:
                    self.log(id(cursor), "CURSORFROM:", "%x" % (id(cursor.getconnection()),),
                             "DB:", self.formatstring(cursor.getconnection().filename, checkmaxlen=False))
        if self.options.sql:
            args=[id(cursor), "SQL:", self.formatstring(sql, '', False)]
            if bindings:
                args.extend(["BINDINGS:", self.format(bindings)])
            self.log(*args)
        return True

    def rowtracer(self, cursor, row):
        if self.options.report:
            self.rowsreturned+=1
        if self.options.rows:
            self.log(id(cursor), "ROW:", self.format(row))
        return row

    def flagme(self, value, mapping, strip=""):
        v=[(k,v) for k,v in mapping.items() if isinstance(k, int)]
        v.sort()
        op=[]
        for k,v in v:
            if value&k:
                if v.startswith(strip):
                    v=v[len(strip):]
                op.append(v)
        return self.u+"|".join(op)

    def connection_hook(self, con):
        self.numconnections+=1
        if self.options.report:
            con.setprofile(self.profiler)
        if self.options.sql or self.options.report:
            con.setexectrace(self.exectracer)
        if self.options.rows or self.options.report:
            con.setrowtrace(self.rowtracer)
        if self.options.sql:
            self.log(id(con), "OPEN:", self.formatstring(con.filename, checkmaxlen=False), con.open_vfs, self.flagme(con.open_flags, self.mapping_open_flags, "SQLITE_OPEN_"))

    def log(self, lid, ltype, *args):
        out=["%x" % (lid,)]
        if self.options.timestamps:
            out.append("%.03f" % (time.time()-self.timestart,))
        if self.options.thread:
            out.append("%x" % (self.threadid(),))
        out.append(ltype)
        out.extend(args)
        self.writer(self.u+" ".join(out))

    def run(self):
        import sys
        import __main__
        d=__main__.__dict__
        if sys.version_info<(3,):
            execfile(sys.argv[0],d, d)
        else:
            # We use compile so that filename is present in printed exceptions
            code=compile(open(sys.argv[0], "rb").read(), sys.argv[0], "exec")
            exec(code, d, d)

    def mostpopular(self, howmany):
        all=[(v,k) for k,v in self.queries.items()]
        all.sort()
        all.reverse()
        return all[:howmany]
        
    def longestrunningaggregate(self, howmany):
        all=[(sum(v),len(v),k) for k,v in self.timings.items()]
        all.sort()
        all.reverse()
        return all[:howmany]

    def longestrunningindividual(self, howmany):
        res=[]
        for k,v in self.timings.items():
            for t in v:
                res.append( (t, k) )
        res.sort()
        res.reverse()
        res=res[:howmany]
        return res

    def report(self):
        import time
        if not self.options.report:
            return
        w=lambda *args: self.writer(self.u+" ".join(args))
        if "summary" in self.options.reports:
            w("APSW TRACE SUMMARY REPORT")
            w()
            w("Program run time                   ", "%.03f seconds" % (time.time()-self.timestart,))
            w("Total connections                  ", str(self.numconnections))
            w("Total cursors                      ", str(self.numcursors))
            w("Number of threads used for queries ", str(len(self.threadsused)))
        total=0
        for k,v in self.queries.items():
            total+=v
        fmtq=len("%d" % (total,))+1
        if "summary" in self.options.reports:
            w("Total queries                      ", str(total))
            w("Number of distinct queries         ", str(len(self.queries)))
            w("Number of rows returned            ", str(self.rowsreturned))
            total=0
            for k,v in self.timings.items():
                for v2 in v:
                    total+=v2
            w("Time spent processing queries      ", "%.03f seconds" % (total/1000000000.0))

        # show most popular queries
        if "popular" in self.options.reports:
            w()
            w("MOST POPULAR QUERIES")
            w()
            for count, query in self.mostpopular(self.options.reportn):
                w("% *d" % (fmtq, count,), self.formatstring(query, '', False))

        # show longest running (aggregate)
        if "aggregate" in self.options.reports:
            w()
            w("LONGEST RUNNING - AGGREGATE")
            w()
            fmtt=None
            for total, count, query in self.longestrunningaggregate(self.options.reportn):
                if fmtt is None:
                    fmtt=len(fmtfloat(total/1000000000.0))+1
                w("% *d %s" % (fmtq, count, fmtfloat(total/1000000000.0, total=fmtt)), self.formatstring(query, '', False))

        # show longest running (individual)
        if "individual" in self.options.reports:
            w()
            w("LONGEST RUNNING - INDIVIDUAL")
            w()
            fmtt=None
            for t,query in self.longestrunningindividual(self.options.reportn):
                if fmtt is None:
                    fmtt=len(fmtfloat(total/1000000000.0))+1
                w(fmtfloat(t/1000000000.0, total=fmtt), self.formatstring(query, '', False))

def fmtfloat(n, decimals=3, total=None):
    "Work around borken python float formatting"
    s="%0.*f" % (decimals, n)
    if total:
        s=(" "*total+s)[-total:]
    return s

def main():
    import optparse
    import os
    import sys

    reports=("summary", "popular", "aggregate", "individual")

    parser=optparse.OptionParser(usage="%prog [options] pythonscript.py [pythonscriptoptions]",
                                 description="This script runs a Python program that uses APSW "
                                 "and reports on SQL queries without modifying the program.  This is "
                                 "done by using connection_hooks and registering row and execution "
                                 "tracers.  See APSW documentation for more details on the output.")
    parser.add_option("-o", "--output", dest="output", default="stdout",
                      help="Where to send the output.  Use a filename, a single dash for stdout, or the words stdout and stderr. [%default]")
    parser.add_option("-s", "--sql", dest="sql", default=False, action="store_true",
                      help="Log SQL statements as they are executed. [%default]")
    parser.add_option("-r", "--rows", dest="rows", default=False, action="store_true",
                      help="Log returned rows as they are returned (turns on sql). [%default]")
    parser.add_option("-t", "--timestamps", dest="timestamps", default=False, action="store_true",
                      help="Include timestamps in logging")
    parser.add_option("-i", "--thread", dest="thread", default=False, action="store_true",
                      help="Include thread id in logging")
    parser.add_option("-l", "--length", dest="length", default=30, type="int",
                      help="Max amount of a string to print [%default]")
    parser.add_option("--no-report", dest="report", default=True, action="store_false",
                      help="A summary report is normally generated at program exit.  This turns off the report and saves memory.")
    parser.add_option("--report-items", dest="reportn", metavar="N", default=15, type="int",
                      help="How many items to report in top lists [%default]")
    parser.add_option("--reports", dest="reports", default=",".join(reports),
                      help="Which reports to show [%default]")

    parser.disable_interspersed_args()
    options, args=parser.parse_args()

    options.reports=[x.strip() for x in options.reports.split(",") if x.strip()]
    for r in options.reports:
        if r not in reports:
            parser.error(r+" is not a valid report.  You should supply one or more of "+", ".join(reports))

    if options.rows:
        options.sql=True

    if not args:
        parser.error("You must specify a python script to execute")

    if not os.path.exists(args[0]):
        parser.error("Unable to find script %r\n" % (args[0],))

    sys.argv=args
    sys.path[0]=os.path.split(os.path.abspath(sys.argv[0]))[0]

    t=APSWTracer(options)

    try:
        t.run()
    finally:
        t.report()

if __name__=="__main__":
    main()
